@@ -1,7 +1,7 @@
 /* Trampoline construction */
 
 /*
- * Copyright 1995-2017 Bruno Haible <bruno@clisp.org>
+ * Copyright 1995-2021 Bruno Haible <bruno@clisp.org>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,20 +14,27 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 
 #include "config.h"
 #include "trampoline_r.h"
 
-#if defined(__hppa__)
+#if defined(__hppa__) && !defined(__hppa64__)
 #if 0
 #define __hppaold__  /* Old trampoline, real machine code. */
 #else
 #define __hppanew__  /* New trampoline, just a closure. */
 #endif
 #endif
+#if defined(__hppa64__)
+#if 0
+#define __hppa64old__  /* Old trampoline, real machine code. */
+#else
+#define __hppa64new__  /* New trampoline, just a closure. */
+#endif
+#endif
 #if defined(__powerpc__) && !defined(__powerpc64__)
 #if !defined(_AIX)
 #define __powerpcsysv4__  /* SysV.4 ABI, real machine code. */
@@ -38,10 +45,10 @@
 #if defined(__powerpc64__) && !defined(__powerpc64_elfv2__)
 #define __powerpc64aix__  /* AIX ABI, just a closure. */
 #endif
-#if defined(__hppanew__)
+#if defined(__hppanew__) || defined(__hppa64new__)
 /*
  * A function pointer is a biased pointer to a data area whose first word
- * contains the actual address of the function.
+ * (hppa) or third word (hppa64) contains the actual address of the function.
  */
 extern void tramp_r (); /* trampoline prototype */
 /* We don't need to take any special measures to make the code executable
@@ -65,28 +72,27 @@ extern void (*tramp_r) (); /* trampoline prototype */
 #endif
 #endif
 
-#ifndef CODE_EXECUTABLE
+#ifndef CODE_EXECUTABLE                 /* Linux, Hurd, macOS, *BSD, IRIX, Solaris, Cygwin, native Windows */
   /* How do we make the trampoline's code executable? */
-  #if defined(HAVE_MACH_VM) || defined(HAVE_WORKING_MPROTECT)
-    #if HAVE_MPROTECT_AFTER_MALLOC_CAN_EXEC > 0
+  #if HAVE_WORKING_MPROTECT             /* Linux, Hurd, macOS, *BSD, IRIX, Solaris, Cygwin */
+    #if HAVE_MPROTECT_AFTER_MALLOC_CAN_EXEC > 0    /* Hurd, macOS, *BSD except HardenedBSD, IRIX, Solaris, Cygwin */
       /* mprotect() [or equivalent] the malloc'ed area. */
       #define EXECUTABLE_VIA_MALLOC_THEN_MPROTECT
     #elif HAVE_MPROTECT_AFTER_MMAP_CAN_EXEC > 0
       /* mprotect() [or equivalent] the mmap'ed area. */
       #define EXECUTABLE_VIA_MMAP_THEN_MPROTECT
-    #elif defined(HAVE_MMAP_SHARED_CAN_EXEC)
+    #elif HAVE_MMAP_SHARED_CAN_EXEC                /* Linux, HardenedBSD */
       #define EXECUTABLE_VIA_MMAP_FILE_SHARED
     #else
       #error "Don't know how to make memory pages executable."
     #endif
-  #else
-    #ifdef HAVE_MMAP
+  #else                                            /* native Windows, very old Linux */
+    #if HAVE_MMAP_ANONYMOUS                        /* very old Linux */
       /* Use an mmap'ed page. */
       #define EXECUTABLE_VIA_MMAP
     #else
-      #ifdef HAVE_SHM
-        /* Use an shmat'ed page. */
-        #define EXECUTABLE_VIA_SHM
+      #if defined(_WIN32) && ! defined(__CYGWIN__) /* native Windows */
+        #define EXECUTABLE_VIA_VIRTUALALLOC
       #else
         ??
       #endif
@@ -99,75 +105,49 @@ extern void (*tramp_r) (); /* trampoline prototype */
 #include <sys/types.h>
 #include <stdlib.h> /* declares abort(), malloc(), free() */
 #ifdef HAVE_UNISTD_H
-#include <unistd.h>
+#include <unistd.h> /* declares getpagesize() */
 #endif
 
+/* Define intptr_t, uintptr_t. */
+#include <stdint.h>
+
 /* Declare getpagesize(). */
-#ifdef HAVE_GETPAGESIZE
+/* On HP-UX, getpagesize exists, but it is not declared in <unistd.h> even if
+   the compiler options -D_HPUX_SOURCE -D_XOPEN_SOURCE=600 are used.  */
+#ifdef __hpux
+extern
 #ifdef __cplusplus
-extern "C" RETGETPAGESIZETYPE getpagesize (void);
-#else
-extern RETGETPAGESIZETYPE getpagesize (void);
+       "C"
 #endif
-#else
-#ifdef HAVE_SYS_PARAM_H
-#include <sys/param.h>
-#else
-/* Not Unix, e.g. mingw32 */
-#define PAGESIZE 4096
-#endif
-#define getpagesize() PAGESIZE
+       int getpagesize (void);
 #endif
 
-/* Declare mprotect() or equivalent. */
+/* Declare mprotect(). */
 #if defined(EXECUTABLE_VIA_MALLOC_THEN_MPROTECT) || defined(EXECUTABLE_VIA_MMAP_THEN_MPROTECT)
-#ifdef HAVE_MACH_VM
-#include <sys/resource.h>
-#include <mach/mach_interface.h>
-#ifdef NeXT
-#include <mach/mach_init.h>
-#endif
-#ifdef __osf__
-#include <mach_init.h>
-#endif
-#include <mach/machine/vm_param.h>
-#else
 #include <sys/types.h>
 #include <sys/mman.h>
 #endif
-#endif
 
 /* Declare mmap(). */
 #if defined(EXECUTABLE_VIA_MMAP) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 #include <sys/types.h>
 #include <sys/mman.h>
-#if !defined(PROT_EXEC) && defined(PROT_EXECUTE) /* Irix 4.0.5 needs this */
-#define PROT_EXEC PROT_EXECUTE
-#endif
 #endif
 
 /* Declare open(). */
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
+#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 #include <sys/types.h>
 #include <unistd.h>
 #include <fcntl.h>
 #endif
 
-/* Declare shmget(), shmat(), shmctl(). */
-#ifdef EXECUTABLE_VIA_SHM
-#include <sys/types.h>
-#include <sys/ipc.h>
-#include <sys/shm.h>
-#ifdef HAVE_SYS_SYSMACROS_H
-#include <sys/sysmacros.h>
-#endif
+/* Declare VirtualAlloc(), GetSystemInfo. */
+#ifdef EXECUTABLE_VIA_VIRTUALALLOC
+#define WIN32_LEAN_AND_MEAN
+#define WIN32_EXTRA_LEAN
+#include <windows.h>
 #endif
 
-/* Some old mmap() implementations require the flag MAP_FILE whenever you pass
-   an fd >= 0. */
-#ifndef MAP_FILE
-#define MAP_FILE 0
-#endif
 /* Some old mmap() implementations require the flag MAP_VARIABLE whenever you
    pass an addr == NULL. */
 #ifndef MAP_VARIABLE
@@ -175,38 +155,52 @@ extern RETGETPAGESIZETYPE getpagesize (void);
 #endif
 
 /* Support for instruction cache flush. */
+#if defined _WIN32 && !defined __CYGWIN__
+# define WIN32_LEAN_AND_MEAN
+# define WIN32_EXTRA_LEAN
+# include <windows.h>
+#elif defined __APPLE__ && defined __MACH__
+# include <libkern/OSCacheControl.h>
+#elif defined _AIX
+# include <sys/cache.h>
+#elif defined __sgi
+# include <sys/cachectl.h>
+#else
 #ifdef __i386__
-#if defined(_WIN32) /* WindowsNT or Windows95 */
+#if defined(_WIN32) && ! defined(__CYGWIN__) /* native Windows */
 #define WIN32_LEAN_AND_MEAN
 #define WIN32_EXTRA_LEAN
 #include <windows.h>
 #endif
 #endif
-#if defined(__mips__) || defined(__mipsn32__) || defined(__mips64__)
-#ifdef linux
-#include <asm/cachectl.h>
+#if defined(__mips__) || defined(__mipsn32__) || defined(__mips64__) || defined(__riscv32__) || defined(__riscv64__)
+#ifdef HAVE_SYS_CACHECTL_H /* IRIX, Linux */
+#include <sys/cachectl.h>
 #else
 #ifdef __OpenBSD__
 #include <machine/sysarch.h>
-#else
-#ifdef HAVE_SYS_CACHECTL_H
-#include <sys/cachectl.h>
-#endif
 #endif
 #endif
 #endif
 /* Inline assembly function for instruction cache flush. */
-#if defined(__sparc__) || defined(__sparc64__) || defined(__alpha__) || defined(__hppaold__) || defined(__powerpcsysv4__) || defined(__powerpc64_elfv2__)
+#if defined(__sparc__) || defined(__sparc64__) || defined(__alpha__) || defined(__hppaold__) || defined(__hppa64old__) || defined(__powerpcsysv4__) || defined(__powerpc64_elfv2__)
 #if defined(__sparc__) || defined(__sparc64__)
 extern void __TR_clear_cache_2();
 #else
 extern void __TR_clear_cache();
 #endif
 #endif
+#endif
 
 /* Support for multithread-safe coding. */
 #include "glthread/lock.h"
 
+#if defined(__APPLE__) && defined(__MACH__) /* only needed on macOS */
+#define KEEP_TEMP_FILE_VISIBLE
+/* Support for temporary files that are cleaned up automatically. */
+#include "clean-temp-simple.h"
+#endif
+
 #if !defined(CODE_EXECUTABLE) && defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 /* Opens a file descriptor and attempts to make it non-inheritable. */
 static int open_noinherit (const char *filename, int flags, int mode)
@@ -270,6 +264,15 @@ static int open_noinherit (const char *filename, int flags, int mode)
 #define TRAMP_ALIGN 16
 #define TRAMP_BIAS 2
 #endif
+#ifdef __hppa64old__
+#define TRAMP_LENGTH 80
+#define TRAMP_ALIGN 8
+#define TRAMP_BIAS 48
+#endif
+#ifdef __hppa64new__
+#define TRAMP_LENGTH 48
+#define TRAMP_ALIGN 8
+#endif
 #if defined(__arm__) || defined(__armhf__)
 #define TRAMP_LENGTH 48
 #define TRAMP_ALIGN 4
@@ -315,6 +318,14 @@ static int open_noinherit (const char *filename, int flags, int mode)
 #define TRAMP_LENGTH 32
 #define TRAMP_ALIGN 8
 #endif
+#ifdef __riscv32__
+#define TRAMP_LENGTH 24
+#define TRAMP_ALIGN 4
+#endif
+#ifdef __riscv64__
+#define TRAMP_LENGTH 32
+#define TRAMP_ALIGN 8
+#endif
 
 #ifndef TRAMP_BIAS
 #define TRAMP_BIAS 0
@@ -326,38 +337,40 @@ static int open_noinherit (const char *filename, int flags, int mode)
 static long pagesize = 0;
 #endif
 
-#if !defined(CODE_EXECUTABLE) && (defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED))
+#if !defined(CODE_EXECUTABLE) && defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 
 /* Variables needed for obtaining memory pages via mmap(). */
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO)
-static int zero_fd;
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
 static int file_fd;
 static long file_length;
-#endif
 
 /* Initialization of these variables. */
 static void for_mmap_init (void)
 {
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO)
-  zero_fd = open("/dev/zero",O_RDONLY,0644);
-  if (zero_fd < 0)
-    { fprintf(stderr,"trampoline: Cannot open /dev/zero!\n"); abort(); }
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
   {
     char filename[100];
     sprintf(filename, "%s/trampdata-%d-%ld", "/tmp", getpid (), random ());
+#if defined(KEEP_TEMP_FILE_VISIBLE)
+    if (register_temporary_file(filename) < 0)
+      { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
+#endif
     file_fd = open_noinherit (filename, O_CREAT | O_RDWR | O_TRUNC, 0700);
     if (file_fd < 0)
-      { fprintf(stderr,"trampoline: Cannot open %s!\n",filename); abort(); }
+      {
+#if defined(KEEP_TEMP_FILE_VISIBLE)
+        unregister_temporary_file(filename);
+#endif
+        fprintf(stderr,"trampoline: Cannot open %s!\n",filename);
+        abort();
+      }
+#if !defined(KEEP_TEMP_FILE_VISIBLE)
     /* Remove the file from the file system as soon as possible, to make
-       sure there is no leftover after this process terminates or crashes. */
+       sure there is no leftover after this process terminates or crashes.
+       On macOS 11.2, this does not work: It would make the mmap call below,
+       with arguments PROT_READ|PROT_EXEC and MAP_SHARED, fail. */
     unlink(filename);
+#endif
   }
   file_length = 0;
-#endif
 }
 
 /* Once-only initializer for these variables. */
@@ -386,17 +399,22 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   /* First, get the page size once and for all. */
   if (!pagesize)
     {
-      /* Simultaneous execution of this initialization in multiple threads
-         is OK. */
-#if defined(HAVE_MACH_VM)
-      pagesize = vm_page_size;
-#else
-      pagesize = getpagesize();
-#endif
-#if defined(EXECUTABLE_VIA_MMAP_DEVZERO) || defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
+#if defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
       /* Use a once-only initializer here, since simultaneous execution of
          for_mmap_init() in multiple threads must be avoided. */
       gl_once (for_mmap_once, for_mmap_init);
+#endif
+      /* Simultaneous execution of this initialization in multiple threads
+         is OK. */
+#if defined(EXECUTABLE_VIA_VIRTUALALLOC)
+      /* GetSystemInfo
+         <https://msdn.microsoft.com/en-us/library/ms724381.aspx>
+         <https://msdn.microsoft.com/en-us/library/ms724958.aspx> */
+      SYSTEM_INFO info;
+      GetSystemInfo(&info);
+      pagesize = info.dwPageSize;
+#else
+      pagesize = getpagesize();
 #endif
     }
 #endif
@@ -409,6 +427,15 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
     { /* Get a new page. */
       char* page;
       char* page_end;
+#ifdef EXECUTABLE_VIA_VIRTUALALLOC
+      /* VirtualAlloc
+         <https://msdn.microsoft.com/en-us/library/aa366887.aspx>
+         <https://msdn.microsoft.com/en-us/library/aa366786.aspx> */
+      page = VirtualAlloc(NULL,pagesize,MEM_COMMIT,PAGE_EXECUTE_READWRITE);
+      if (page == NULL)
+        { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
+      page_end = page + pagesize;
+#else
 #ifdef EXECUTABLE_VIA_MMAP_FILE_SHARED
       char* page_x;
       /* Extend the file by one page. */
@@ -423,43 +450,27 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
       file_length = new_file_length;
       page_end = page + pagesize;
       /* Link the two pages together. */
-      ((long*)page)[0] = page_x - page;
-      page = (char*)(((long)page + sizeof(long) + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
+      ((intptr_t*)page)[0] = page_x - page;
+      page = (char*)(((uintptr_t)page + sizeof(intptr_t) + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
 #else
 #ifdef EXECUTABLE_VIA_MMAP_THEN_MPROTECT
-#ifdef HAVE_MMAP_ANONYMOUS
       /* Use mmap with the MAP_ANONYMOUS or MAP_ANON flag. */
       page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS | MAP_VARIABLE, -1, 0);
-#else
-      /* Use mmap on /dev/zero. */
-      page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_FILE | MAP_VARIABLE, zero_fd, 0);
-#endif
 #endif
 #ifdef EXECUTABLE_VIA_MMAP
-#ifdef HAVE_MMAP_ANONYMOUS
       /* Use mmap with the MAP_ANONYMOUS or MAP_ANON flag. */
       page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE | PROT_EXEC, MAP_PRIVATE | MAP_ANONYMOUS | MAP_VARIABLE, -1, 0);
-#else
-      /* Use mmap on /dev/zero. */
-      page = mmap(NULL, pagesize, PROT_READ | PROT_WRITE | PROT_EXEC, MAP_PRIVATE | MAP_FILE | MAP_VARIABLE, zero_fd, 0);
-#endif
-#endif
-#ifdef EXECUTABLE_VIA_SHM
-      int shmid = shmget(IPC_PRIVATE, pagesize, 0700|IPC_CREAT);
-      if (shmid<0)
-        { page = (char*)(-1); }
-      else
-        { page = shmat(shmid, 0, 0); shmctl(shmid, IPC_RMID, 0); }
 #endif
       if (page == (char*)(-1))
         { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
       page_end = page + pagesize;
+#endif
 #endif
       /* Fill it with free trampolines. */
       { char** last = &freelist;
         while (page+TRAMP_TOTAL_LENGTH <= page_end)
           { *last = page; last = (char**)page;
-            page = (char*)(((long)page + TRAMP_TOTAL_LENGTH + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
+            page = (char*)(((uintptr_t)page + TRAMP_TOTAL_LENGTH + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
           }
         *last = NULL;
     } }
@@ -469,15 +480,15 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   { char* room = (char*) malloc(sizeof(void*) + TRAMP_TOTAL_LENGTH + TRAMP_ALIGN-1);
     if (!room)
       { fprintf(stderr,"trampoline: Out of virtual memory!\n"); abort(); }
-    function = (char*)(((long)room + sizeof(void*) + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
+    function = (char*)(((uintptr_t)room + sizeof(void*) + TRAMP_ALIGN-1) & -TRAMP_ALIGN);
     ((char**)function)[-1] = room; /* backpointer for free_trampoline() */
   }
 #endif
 
 #if !defined(CODE_EXECUTABLE) && defined(EXECUTABLE_VIA_MMAP_FILE_SHARED)
   /* Find the executable address corresponding to the writable address. */
-  { unsigned long page = (unsigned long) function & -(long)pagesize;
-    function_x = function + ((long*)page)[0];
+  { uintptr_t page = (uintptr_t) function & -(intptr_t)pagesize;
+    function_x = function + ((intptr_t*)page)[0];
   }
 #else
   function_x = function;
@@ -534,14 +545,14 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   /* function:
    *    lw $2,16($25)			8F 22 00 10
    *    lw $25,20($25)			8F 39 00 14
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    *    .word <data>			<data>
    *    .word <address>			<address>
    */
   *(unsigned int *) (function + 0) = 0x8F220010;
   *(unsigned int *) (function + 4) = 0x8F390014;
-  *(unsigned int *) (function + 8) = 0x03200008;
+  *(unsigned int *) (function + 8) = 0x03200009;
   *(unsigned int *) (function +12) = 0x00000000;
   *(unsigned int *) (function +16) = (unsigned int) data;
   *(unsigned int *) (function +20) = (unsigned int) address;
@@ -549,7 +560,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #define is_tramp(function)  \
   *(unsigned int *) (function + 0) == 0x8F220010 && \
   *(unsigned int *) (function + 4) == 0x8F390014 && \
-  *(unsigned int *) (function + 8) == 0x03200008 && \
+  *(unsigned int *) (function + 8) == 0x03200009 && \
   *(unsigned int *) (function +12) == 0x00000000
 #define tramp_address(function)  \
   *(unsigned int *) (function +20)
@@ -570,7 +581,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
    *					37 39 hi16(lo32(<address>))
    *					00 19 CC 38
    *					37 39 lo16(lo32(<address>))
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    */
   /* What about big endian / little endian ?? */
@@ -594,7 +605,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   *(int *)   (function +40) = 0x0019CC38;
   *(short *) (function +44) = 0x3739;
   *(short *) (function +46) = (unsigned long) address & 0xffff;
-  *(int *)   (function +48) = 0x03200008;
+  *(int *)   (function +48) = 0x03200009;
   *(int *)   (function +52) = 0x00000000;
 #define is_tramp(function)  \
   *(unsigned short *) (function + 0) == 0x3C02 && \
@@ -609,7 +620,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   *(unsigned short *) (function +36) == 0x3739 && \
   *(unsigned int *)   (function +40) == 0x0019CC38 && \
   *(unsigned short *) (function +44) == 0x3739 && \
-  *(unsigned int *)   (function +48) == 0x03200008 && \
+  *(unsigned int *)   (function +48) == 0x03200009 && \
   *(unsigned int *)   (function +52) == 0x00000000
 #define hilo(word3,word2,word1,word0)  \
   (((unsigned long) (word3) << 48) | ((unsigned long) (word2) << 32) | \
@@ -629,14 +640,14 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   /* function:
    *    ld $2,16($25)			DF 22 00 10
    *    ld $25,24($25)			DF 39 00 18
-   *    j $25				03 20 00 08
+   *    jal $0,$25			03 20 00 09  was:  j $25   03 20 00 08
    *    nop				00 00 00 00
    *    .dword <data>			<data>
    *    .dword <address>		<address>
    */
   *(unsigned int *)  (function + 0) = 0xDF220010;
   *(unsigned int *)  (function + 4) = 0xDF390018;
-  *(unsigned int *)  (function + 8) = 0x03200008;
+  *(unsigned int *)  (function + 8) = 0x03200009;
   *(unsigned int *)  (function +12) = 0x00000000;
   *(unsigned long *) (function +16) = (unsigned long) data;
   *(unsigned long *) (function +24) = (unsigned long) address;
@@ -644,7 +655,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #define is_tramp(function)  \
   *(unsigned int *)  (function + 0) == 0xDF220010 && \
   *(unsigned int *)  (function + 4) == 0xDF390018 && \
-  *(unsigned int *)  (function + 8) == 0x03200008 && \
+  *(unsigned int *)  (function + 8) == 0x03200009 && \
   *(unsigned int *)  (function +12) == 0x00000000
 #define tramp_address(function)  \
   *(unsigned long *) (function +24)
@@ -822,6 +833,76 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #define tramp_data(function)  \
   ((long *) function)[2]
 #endif
+#ifdef __hppa64old__
+  /* function:
+   *    mfia %r27			000014BB
+   *    ldd 32(%r27),%r31		537F0040
+   *    ldd 40(%r27),%r27		537B0050
+   *    ldd 16(%r27),%r1		53610020
+   *    ldd 24(%r27),%r27		537B0030
+   *    bve (%r1)			E820D000
+   *     nop				08000240
+   *    .align 8
+   *    .dword <data>
+   *    .dword <address>
+   * function_pointer:
+   *    .dword 0
+   *    .dword 0
+   *    .dword function
+   *    .dword 0
+   */
+  *(int *)  (function + 0) = 0x000014BB;
+  *(int *)  (function + 4) = 0x537F0040;
+  *(int *)  (function + 8) = 0x537B0050;
+  *(int *)  (function +12) = 0x53610020;
+  *(int *)  (function +16) = 0x537B0030;
+  *(int *)  (function +20) = 0xE820D000;
+  *(int *)  (function +24) = 0x08000240;
+  *(long *) (function +32) = (long)data;
+  *(long *) (function +40) = (long)address;
+  *(long *) (function +48) = (long)0;
+  *(long *) (function +56) = (long)0;
+  *(long *) (function +64) = (long)function;
+  *(long *) (function +72) = (long)0;
+#define TRAMP_CODE_LENGTH  28
+#define is_tramp(function)  \
+  *(int *) (function + 0) == 0x000014BB && \
+  *(int *) (function + 4) == 0x537F0040 && \
+  *(int *) (function + 8) == 0x537B0050 && \
+  *(int *) (function +12) == 0x53610020 && \
+  *(int *) (function +16) == 0x537B0030 && \
+  *(int *) (function +20) == 0xE820D000 && \
+  *(int *) (function +24) == 0x08000240
+#define tramp_address(function)  \
+  (*(unsigned long *) (function +40))
+#define tramp_data(function)  \
+  (*(unsigned long *) (function +32))
+#endif
+#ifdef __hppa64new__
+  /* function:
+   *    .dword 0
+   *    .dword 0
+   *    .dword tramp
+   *    .dword closure
+   * closure:
+   *    .dword <data>
+   *    .dword <address>
+   */
+  *(long *) (function + 0) = 0;
+  *(long *) (function + 8) = 0;
+  *(long *) (function +16) = ((long *) (void*) &tramp_r)[2];
+  *(long *) (function +24) = (long) (function + 32);
+  *(long *) (function +32) = (long) data;
+  *(long *) (function +40) = (long) address;
+#define TRAMP_CODE_LENGTH  0
+#define is_tramp(function)  \
+  ((long *) function)[2] == ((long *) (void*) &tramp_r)[2] && \
+  ((long *) function)[3] == (long) (function + 32)
+#define tramp_address(function)  \
+  ((long *) function)[5]
+#define tramp_data(function)  \
+  ((long *) function)[4]
+#endif
 #if defined(__arm__) || defined(__armhf__)
   /* function:
    *	mov	ip,sp			E1A0C00D
@@ -1049,22 +1130,22 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
    *    jmp *%rax			FF E0
    */
   *(short *) (function + 0) = 0xBA49;
-  *(short *) (function + 2) = (unsigned long) data & 0xffff;
-  *(int *)   (function + 4) = ((unsigned long) data >> 16) & 0xffffffff;
-  *(short *) (function + 8) = ((unsigned long) data >> 48) & 0xffff;
+  *(short *) (function + 2) = (unsigned long long) data & 0xffff;
+  *(int *)   (function + 4) = ((unsigned long long) data >> 16) & 0xffffffff;
+  *(short *) (function + 8) = ((unsigned long long) data >> 48) & 0xffff;
   *(short *) (function +10) = 0xB848;
-  *(int *)   (function +12) = (unsigned long) address & 0xffffffff;
-  *(int *)   (function +16) = ((unsigned long) address >> 32) & 0xffffffff;
+  *(int *)   (function +12) = (unsigned long long) address & 0xffffffff;
+  *(int *)   (function +16) = ((unsigned long long) address >> 32) & 0xffffffff;
   *(short *) (function +20) = 0xE0FF;
 #define is_tramp(function)  \
   *(unsigned short *) (function + 0) == 0xBA49 && \
   *(unsigned short *) (function +10) == 0xB848 && \
   *(unsigned short *) (function +20) == 0xE0FF
 #define hilo(hiword,loword)  \
-  (((unsigned long) (hiword) << 32) | (unsigned long) (loword))
+  (((unsigned long long) (hiword) << 32) | (unsigned long long) (loword))
 #define himidlo(hishort,midword,loshort)  \
-  (((unsigned long) (hishort) << 48) | (unsigned long) (midword) << 16 \
-   | (unsigned long) (loshort))
+  (((unsigned long long) (hishort) << 48) | (unsigned long long) (midword) << 16 \
+   | (unsigned long long) (loshort))
 #define tramp_address(function)  \
   hilo(*(unsigned int *) (function +16), *(unsigned int *) (function +12))
 #define tramp_data(function)  \
@@ -1124,30 +1205,78 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   (*(unsigned long *) (function +24))
 #define tramp_data(function)  \
   (*(unsigned long *) (function +16))
+#endif
+#ifdef __riscv32__
+  /* function:
+   *    auipc t0,0			00000297
+   *    lw t1,20(t0)			0142A303
+   *    lw t2,16(t0)			0102A383
+   *    jr t1				00030067
+   * data:    .quad <data>
+   * address: .quad <address>
+   */
+  *(int *)   (function + 0) = 0x00000297;
+  *(int *)   (function + 4) = 0x0142A303;
+  *(int *)   (function + 8) = 0x0102A383;
+  *(int *)   (function +12) = 0x00030067;
+  *(int *)   (function +16) = (unsigned int) data;
+  *(int *)   (function +24) = (unsigned int) address;
+#define TRAMP_CODE_LENGTH  16
+#define is_tramp(function)  \
+  *(unsigned int *) (function + 0) == 0x00000297 && \
+  *(unsigned int *) (function + 4) == 0x0142A303 && \
+  *(unsigned int *) (function + 8) == 0x0102A383 && \
+  *(unsigned int *) (function +12) == 0x00030067
+#define tramp_address(function)  \
+  (*(unsigned int *) (function +20))
+#define tramp_data(function)  \
+  (*(unsigned int *) (function +16))
+#endif
+#ifdef __riscv64__
+  /* function:
+   *    auipc t0,0			00000297
+   *    ld t1,24(t0)			0182B303
+   *    ld t2,16(t0)			0102B383
+   *    jr t1				00030067
+   * data:    .quad <data>
+   * address: .quad <address>
+   */
+  *(int *)   (function + 0) = 0x00000297;
+  *(int *)   (function + 4) = 0x0182B303;
+  *(int *)   (function + 8) = 0x0102B383;
+  *(int *)   (function +12) = 0x00030067;
+  *(long *)  (function +16) = (unsigned long) data;
+  *(long *)  (function +24) = (unsigned long) address;
+#define TRAMP_CODE_LENGTH  16
+#define is_tramp(function)  \
+  *(unsigned int *) (function + 0) == 0x00000297 && \
+  *(unsigned int *) (function + 4) == 0x0182B303 && \
+  *(unsigned int *) (function + 8) == 0x0102B383 && \
+  *(unsigned int *) (function +12) == 0x00030067
+#define tramp_address(function)  \
+  (*(unsigned long *) (function +24))
+#define tramp_data(function)  \
+  (*(unsigned long *) (function +16))
 #endif
   /*
    * data:
    *    <data0>				<data0>
    *    <data1>				<data1>
    */
-  *(long *)  (data + 0*sizeof(void*)) = (long) data0;
-  *(long *)  (data + 1*sizeof(void*)) = (long) data1;
+  *(void* *) (data + 0*sizeof(void*)) = data0;
+  *(void* *) (data + 1*sizeof(void*)) = data1;
 
   /* 3. Set memory protection to "executable" */
 
 #if !defined(CODE_EXECUTABLE)
 #if defined(EXECUTABLE_VIA_MALLOC_THEN_MPROTECT) || defined(EXECUTABLE_VIA_MMAP_THEN_MPROTECT)
   /* Call mprotect on the pages that contain the range. */
-  { unsigned long start_addr = (unsigned long) function;
-    unsigned long end_addr = (unsigned long) (function + TRAMP_LENGTH);
+  { uintptr_t start_addr = (uintptr_t) function;
+    uintptr_t end_addr = (uintptr_t) (function + TRAMP_LENGTH);
     start_addr = start_addr & -pagesize;
     end_addr = (end_addr + pagesize-1) & -pagesize;
-   {unsigned long len = end_addr - start_addr;
-#if defined(HAVE_MACH_VM)
-    if (vm_protect(task_self(),start_addr,len,0,VM_PROT_READ|VM_PROT_WRITE|VM_PROT_EXECUTE) != KERN_SUCCESS)
-#else
+   {uintptr_t len = end_addr - start_addr;
     if (mprotect((void*)start_addr, len, PROT_READ|PROT_WRITE|PROT_EXEC) < 0)
-#endif
       { fprintf(stderr,"trampoline: cannot make memory executable\n"); abort(); }
   }}
 #endif
@@ -1162,8 +1291,31 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #ifndef TRAMP_CODE_LENGTH
 #define TRAMP_CODE_LENGTH TRAMP_LENGTH
 #endif
-#if !(defined(__hppanew__) || defined(__powerpcaix__) || defined(__powerpc64aix__) || defined(__ia64__))
+#if !(defined(__hppanew__) || defined(__hppa64new__) || defined(__powerpcaix__) || defined(__powerpc64aix__) || defined(__ia64__))
   /* Only needed if we really set up machine instructions. */
+  /* Use the operating system provided function, when available. */
+#if defined _WIN32 && !defined __CYGWIN__
+  /* Native Windows.
+     FlushInstructionCache
+     <https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-flushinstructioncache>  */
+  HANDLE process = GetCurrentProcess ();
+  while (!FlushInstructionCache (process, function_x, TRAMP_CODE_LENGTH))
+    ;
+#elif defined __APPLE__ && defined __MACH__
+  /* macOS  */
+  sys_icache_invalidate (function_x, TRAMP_CODE_LENGTH);
+#elif defined _AIX
+  /* AIX.  */
+  _sync_cache_range (function_x, TRAMP_CODE_LENGTH);
+#elif defined __sgi
+  /* IRIX.  */
+  cacheflush (function_x, TRAMP_CODE_LENGTH, ICACHE);
+#elif defined __sun
+  /* Solaris.  */
+  extern void sync_instruction_memory (char *, size_t);
+  sync_instruction_memory (function_x, TRAMP_CODE_LENGTH);
+#else
+  /* No operating system provided function. Dispatch according to the CPU. */
 #ifdef __i386__
 #if defined(_WIN32)
   while (!FlushInstructionCache(GetCurrentProcess(),function_x,TRAMP_CODE_LENGTH))
@@ -1172,8 +1324,8 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #endif
 #ifdef __m68k__
 #if defined(__NetBSD__) && defined(__GNUC__)
-  { register unsigned long _beg __asm__ ("%a1") = (unsigned long) function_x;
-    register unsigned long _len __asm__ ("%d1") = TRAMP_CODE_LENGTH;
+  { register uintptr_t _beg __asm__ ("%a1") = (uintptr_t) function_x;
+    register uintptr_t _len __asm__ ("%d1") = TRAMP_CODE_LENGTH;
     __asm__ __volatile__ (
       "move%.l %#0x80000004,%/d0\n\t" /* CC_EXTPURGE | C_IPURGE */
       "trap #12"                      /* kernel call ‘cachectl’ */
@@ -1184,8 +1336,8 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
   }
 #endif
 #if defined(__linux__) && defined(__GNUC__)
-  { register unsigned long _beg __asm__ ("%d1") = (unsigned long) function_x;
-    register unsigned long _len __asm__ ("%d4") = TRAMP_CODE_LENGTH + 32;
+  { register uintptr_t _beg __asm__ ("%d1") = (uintptr_t) function_x;
+    register uintptr_t _len __asm__ ("%d4") = TRAMP_CODE_LENGTH + 32;
     __asm__ __volatile__ (
       "move%.l %#123,%/d0\n\t"
       "move%.l %#1,%/d2\n\t"
@@ -1226,7 +1378,7 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #ifdef __alpha__
   __TR_clear_cache();
 #endif
-#ifdef __hppa__
+#if defined(__hppa__) || defined(__hppa64__)
   /* This assumes that the trampoline fits in at most two cache lines. */
   __TR_clear_cache(function_x,function_x+TRAMP_CODE_LENGTH-1);
 #endif
@@ -1248,6 +1400,15 @@ __TR_function alloc_trampoline_r (__TR_function address, void* data0, void* data
 #if defined(__powerpc__) || defined(__powerpc64__)
   __TR_clear_cache(function_x);
 #endif
+#if defined(__riscv32__) || defined(__riscv64__)
+#if defined(__linux__)
+  /* Use the libc function. */
+  __riscv_flush_icache((void*)function_x,(void*)(function_x+TRAMP_CODE_LENGTH),0);
+#elif defined(__GNUC__)
+  __asm__ __volatile__ ("fence.i");
+#endif
+#endif
+#endif
 #endif
 
   /* 5. Return. */
@@ -1262,11 +1423,13 @@ void free_trampoline_r (__TR_function function)
 #if !defined(CODE_EXECUTABLE) && !defined(EXECUTABLE_VIA_MALLOC_THEN_MPROTECT)
 #ifdef EXECUTABLE_VIA_MMAP_FILE_SHARED
   /* Find the writable address corresponding to the executable address. */
-  { unsigned long page_x = (unsigned long) function & -(long)pagesize;
-    function -= ((long*)page_x)[0];
+  { uintptr_t page_x = (uintptr_t) function & -(intptr_t)pagesize;
+    function -= ((intptr_t*)page_x)[0];
   }
 #endif
+  gl_lock_lock(freelist_lock);
   *(char**)function = freelist; freelist = (char*)function;
+  gl_lock_unlock(freelist_lock);
   /* It is probably not worth calling munmap() for entirely freed pages. */
 #else
   free(((char**)function)[-1]);
@@ -1278,20 +1441,20 @@ int is_trampoline_r (void* function)
 #if defined(is_tramp) && defined(tramp_data)
 #ifdef __hppanew__
   void* tramp_r_address = &tramp_r;
-  if (!(((long)function & 3) == (TRAMP_BIAS & 3))) return 0;
+  if (!(((uintptr_t)function & 3) == (TRAMP_BIAS & 3))) return 0;
 #endif
   if (is_tramp(((char*)function - TRAMP_BIAS)))
     {
       char* function_w;
 #ifdef EXECUTABLE_VIA_MMAP_FILE_SHARED
       /* Find the writable address corresponding to the executable address. */
-      { unsigned long page_x = (unsigned long) function & -(long)pagesize;
-        function_w = function - ((long*)page_x)[0];
+      { uintptr_t page_x = (uintptr_t) function & -(intptr_t)pagesize;
+        function_w = function - ((intptr_t*)page_x)[0];
       }
 #else
       function_w = function;
 #endif
-      return (tramp_data(((char*)function - TRAMP_BIAS))) == (long)((char*)function_w - TRAMP_BIAS + TRAMP_LENGTH);
+      return (tramp_data(((char*)function - TRAMP_BIAS))) == (uintptr_t)((char*)function_w - TRAMP_BIAS + TRAMP_LENGTH);
     }
   return 0;
 #else
@@ -1299,7 +1462,7 @@ int is_trampoline_r (void* function)
 #endif
 }
 
-__TR_function trampoline_r_address (void* function)
+__TR_function trampoline_r_address (__TR_function function)
 {
 #ifdef tramp_address
   return (__TR_function)(tramp_address(((char*)function - TRAMP_BIAS)));
@@ -1308,7 +1471,7 @@ __TR_function trampoline_r_address (void* function)
 #endif
 }
 
-void* trampoline_r_data0 (void* function)
+void* trampoline_r_data0 (__TR_function function)
 {
 #ifdef tramp_data
   return ((void**)((char*)function-TRAMP_BIAS+TRAMP_LENGTH))[0];
@@ -1317,7 +1480,7 @@ void* trampoline_r_data0 (void* function)
 #endif
 }
 
-void* trampoline_r_data1 (void* function)
+void* trampoline_r_data1 (__TR_function function)
 {
 #ifdef tramp_data
   return ((void**)((char*)function-TRAMP_BIAS+TRAMP_LENGTH))[1];
